library(Hmisc)

# calling this auto-sets the R width. 
fix_width <- function(){ options(width=Sys.getenv("COLUMNS")) }


resid <- function(a,b) { residuals(lm(a ~ b, na.action=na.exclude)) }

# A grep for line by line patterns
lgrepl <- function(pat, x) {
	out <- rep(NA, length(x))
	for(i in 1:length(x)) {
		out[i] <- grepl(pat[i], x[i])
	}
	out
}

inv.logit <- function(x) {
	exp(x) / (1+exp(x))
}

#################################################################################
##  Load a directory full of subject files

load_subject_directory <- function(dir, names, pattern=".txt", fixTF=T, add.item.number=F) {
	# Load a directory of subject files, in the same format
	# - pattern - filters only files of that type
	# - fixTF -- should "True" and "False" be converted to "TRUE" and "FALSE

	d <- NULL
	for(f in list.files(dir, pattern=pattern, full.names=T)) {
		print(paste("# Loading subject file", f), quote=F)
		
		q <- read.table(f, header=F)
		names(q) <- names
		q$subject <- f
		
		if(add.item.number){ 
			q$item.number <- cumsum(q$response.number == 1) # how many first responses have we given?
		}

		d <- rbind(d,q)
	}
	
	# and fix the true/false format
	if(fixTF) {
		for(c in 1:ncol(d)) {
			if( length(setdiff(d[,c], c("True", "T", "TRUE", "False", "F", "FALSE", "NA", NA, "<NA>"))) == 0) { # if all are these
				newcol <- is.element(d[,c], c("True", "T", "TRUE"))
				newcol[ is.na(d[,c]) ] <- NA # fix the NAs
				d[,c] <- newcol
			}
		}
	}


	return(d)
}



#################################################################################
##

first <- function(x) { x[1] }

assert <- function(tv, message) { 
	if( ! tv) {
		print(message)
		q()
	}

}


# this makes a polygon between upper and lower (e.g., it wraps lower the right way so it goes back
polygon.ul <- function(x, lower, upper, ...) {
	polygon( c(x, rev(x)), c(upper, rev(lower)), ...)
}

# A special function for binning binomial outcomes and computing
# the error bars for each time point
# stats controls whether or not we get confidence intervals, etc. Otherwise, we just compute mean pct etc
aggregate.binomial <- function(outcome, by, na.rm=T, names=NULL, outcome.name="x", stats=TRUE) {
	
	# k out of n
	k <- aggregate(outcome, by=by, function(x) { sum(x,na.rm=na.rm) })
	n <- aggregate(outcome, by=by, function(x) { length(x[!is.na(x)]) } )
	
	if( !is.null(names) ) { names(k) <- c(names, "x") }
	
	# because we aggregated by the same thing, we can just map the rows
	k$n <- n$x
	k$p <- NA
	k$lower <- NA
	k$upper <- NA
	k$p.value <- NA

	if(stats) {
		
		for(i in 1:dim(n)[1]) {
			
			if(k$n[i] > 0) {
			t <- binom.test(k$x[i], k$n[i])
			k$p[i] <- t$estimate
			k$lower[i] <- t$conf.int[1]
			k$upper[i] <- t$conf.int[2]		
			k$p.value[i] <- t$p.value
			}
		}
	}
	else { # don't run stats, just compute the prob 
		k$p <- k$x / k$n
	}

	# set what we made "x" to be outcome.name
	names(k)[length(names)+1] <-  outcome.name
	
	k
	
}

 #my own version
drop.levels <- function(x) {
	factor(as.character(x))
}


my.aggregate <- function(..., names=c("Group.1", "x")) {
	d <- aggregate(...)
	names(d) <- names
	d
}

# like seq but will take log=TRUE
logseq <- function(from = 1, to = 1, n=10 ) {
	
	exp(seq( log(from), log(to), by=(log(to) - log(from))/n ) )
}

not.na <- function(x) { x[!is.na(x)] }

# removes NA and NaN and Inf
nums <- function(x) {
        x[ !is.na(x) & x < Inf & x > -Inf & !is.nan(x) ]
}

sem <- function(x) { 
        if(length(not.na(x)) < 2) { NA }
        else { sd( not.na(x) ) / sqrt(length( not.na(x) )) }
}

medians <- function(a, b) { c(median(a), median(b)) }
means <- function(a, b) { c(mean(a), mean(b)) }
sds <- function(a, b) { c(sd(a), sd(b)) }

stdize <- function(x, ...) {
        (x - mean(x,...)) / sd(x, ...)
}


# aggregate by two functions (ie mean and sem)
aggregate2 <- function(x, by, FUN1=mean, FUN2=sem, names=c("X", "F1", "F2") ) {
        a <- aggregate(x, by=by, FUN=FUN1)
        a2 <- aggregate(x, by=by, FUN=FUN2)
        a$x2 <- a2$x # the rows should be the same since we aggregated by the same thing
        names(a) <- names
        a
}

plot.eb <- function(x, y, sem, add=FALSE, pch=19, col=1, type="b", xlab="x", ylab="y", ylim="auto", ...) {
        if(add) {
                lines(x,y, pch=pch, col=col, type=type, ...)
                errbar(x, y, y+2*sem, y-2*sem, pch=pch, col=col, add=TRUE)
        }
        else {
		yl <- ylim
		if(length(ylim==1) & ylim == "auto") { yl <- c( min(nums(y-2*sem)), max(nums(y+2*sem))) }
		
                plot(x, y, pch=pch, col=col, ylim=yl, type=type, xlab=xlab, ylab=ylab, ...)
                errbar(x, y, y+2*sem, y-2*sem, pch=pch, col=col, add=TRUE)
        }
}




#################################################################################
## These convert to and from our file name format

# this takes a string and matches everything of the form a __1_2 and returns a
# "hash" of ret[1]=2
params2hash <- function(s) {
	ret <- NULL
	# remove a file exension from s
	s <- sub("\\.txt$", "", s, perl=T)
	
	parts <- strsplit(s, "__+", perl=T)[[1]];
	for(p in parts) {
# 		print(p)
		if(grepl("_", p)) {
			kv = strsplit(p, "_")[[1]];
# 			print(kv[1])
			ret[kv[1]] = kv[2]
		}
	}
	ret	
}

# hash2param converts a list of hashes to a parameter file name
hash2param <- function(pre="", x, post=".txt") {
	
	str <- pre;
	for(y in names(x)) {
		str <- paste(str, "__", y, "_", x[y], sep="");
	}
	str <- paste(str, post, sep="")

	str
}

# take a hash (strings to values) and convert to a string
params2string <- function(x, prefix="", postfix="") {

	str <- prefix
	for(i in 1:length(x) ) {
		str <- paste(str, "__", names(x)[i], "_", x[i],  sep="") 
	}
	str <- paste(str, postfix, sep="");

	str;

}


logit  <- function(x) { log(x/(1.0-x)) }
ilogit <- function(x) { 1.0/(1.0+exp(-x)) }



