

;; This takes care of all of the caching etc
;; and implements:

;; (model-push-hypothesis! h resp) -- pushes the hypothesis h with respones resp onto data
;; (model-set-correct! d) -- a set of data
;; (model-set-which-item! w) -- a list saying which item we are on
;; (model-set-data! d) -- sets the data that this model is run on (for doing exemplar, etc)
;; (model-prediction p) -- takes a parameter value and computes the model predicted proportions (possibly with fancy caching)
;; (get-initial-p) -- what p we can start with?
;; (get-proprosers) -- proposal functions for each of the parameters

; http://community.schemewiki.org/?simple-object


; -----------------------------------------------------------------------------------------
;; Define across parameters (outside here, so that they are shared)

(define (set-model)
  (set! across-parameters (vector 1.0 1.0 1.0 1.0 -0.5))
  (set! across-proposers (vector
                          (lambda (x) (normal-proposal x 0.05)) ;; length weight
                          (lambda (x) (normal-proposal x 0.05)) ;; shape weight
                          (lambda (x) (normal-proposal x 0.05)) ;; size weight
                          (lambda (x) (normal-proposal x 0.05))
                          (lambda (x) (negative-normal-proposal x 0.025))
                          )) ;; color weight
  (set! make-model make-model:exemplar)) ;; color weight
                          
; -----------------------------------------------------------------------------------------
(define (make-model:exemplar)
  
  (define META-description "Simple exemplar model." )
  (define META-notes       "(Uses C external library for running-ll)." )
  
  (define hypotheses null) ;; a vector of hypotheses
  (define responses null) ;; a vector of hypotheses
  (define priors null) ;; a vector of hypotheses
  (define correct null) ;; stored in collapsed format
  (define which-item null)
  (define zdata null) ;; for when we need to store the data
  (define len 0)
  
  (define (initialize-model) 0)  
  
  ; (define (get-initial-p) (vector 1.0 1.0 1.0 1.0 -0.15))
  (define (get-initial-p) (vector))
  (define (get-proposers) (vector))
  
                           
                           
  (define (model-set-correct! c) 
    (set! correct c)
    (set! len (length correct))
    )
  
  (define (model-set-which-item! w) 
    (set! which-item w)
    )
  
  (define (model-set-data! d) ;; this stores zipped versions of the data
    (set! zdata (map (lambda (x) (zip-list x))
                     d)))
  
  ;; adds h to the model, with c and l
  (define (model-push-hypothesis! h resp) 
    (set! hypotheses (cons h hypotheses ))
    (set! responses  (cons resp responses ))
   ; (set! priors     (cons (PCFG-log-probability h) priors))
    )
   
  ;; --------------------------
  ;; Local helper functions
  
  
  ;; This takes a function f which computes the distance and returns the distance consed
  ;; to a permuted version of y such that y aligns with x with the min distance possible
  ;; this returns (distance y1 y2 y3 y4 ...) where the yi are ordered to align with x
  ;; NOTE: Some of the ys may be null when x is larger than y -- this means we skip aligning to some x
  ;; NOTE: This is a factorial time algorithm! Jesus. 
  ;;; TODO: Can make this break out early if it has to
  (define (min-permutation-over f x y)
    
    (if (null? x)
        '(0) ;; this gives a null cdr, which is what we want for the recursion base case
        (let ((best-distance +inf.0)
              (best-order null))
            
          ;; first try mapping (car x) to null, if we have more x than y
          (if (> (length x) (length y))
              (let* ((ret (min-permutation-over f (cdr x) y)) ;; recurse assuming nothing maps to (car x)
                     (dist (first ret)))
                (if (< dist best-distance)
                    (begin (set! best-distance dist)
                           (set! best-order (cons null (cdr ret)))))))
        
          ;; now loop over i (yi maps to car x) and set the bests
          (for-each (lambda (i)
                      (let* ((yi (nth i y))
                             (ret (min-permutation-over f (cdr x) (kill-nth i y))) ;; recurse assuming yi maps to (car x)
                             (dist (+ (f (car x) yi)
                                      (first ret))))
                        
                        (if (< dist best-distance)
                            (begin (set! best-distance dist)
                                   (set! best-order (cons yi (cdr ret)))))
                        
                        ))
                    (seq 1 (length y) 1))
          
          ;; now best-distance stores the best distance and best-order stores the best order (including (car x))
          ;; so just return them
          (cons best-distance best-order))))

  
  ;; --------------------------
  (define (model-prediction p gp ap)
    
    (let ((PARAM-weight-length-difference (vector-ref ap 0))
          (PARAM-weight-shape             (vector-ref ap 1))
          (PARAM-weight-size              (vector-ref ap 2))
          (PARAM-weight-color             (vector-ref ap 3))
          
          ;; fit this within
          (PARAM-llBeta            (vector-ref ap 4))
          ;(PARAM-llBeta-negative            (vector-ref ap 5))
          
          (dlen (length zdata)))
      
      (apply append  
           (formap xi in (seq 1 dlen) ;;for each set in data
                   
                   (let* ((x (nth xi zdata))
                          ;; the summed prob of true and false for each response in x
                          ;; this is filled in via the loop below
                          (respn-T (make-vector (length x) 0.001) ) ;; Hmm start these out with tiny counts for smoothing. 
                          (respn-F (make-vector (length x) 0.001) ))
                     
                     (for yi in (seq 1 (- xi 1))
                       
                       (let* ((y (nth yi zdata))
                              (perm (min-permutation-over (lambda (a b) ;; make a function to compute the distance with these params
                                                            (+ (if (not (eqv-shape* (second a) (second b))) PARAM-weight-shape 0)
                                                               (if (not (eqv-size* (second a) (second b))) PARAM-weight-size 0)
                                                               (if (not (eqv-color* (second a) (second b))) PARAM-weight-color 0)))
                                                          x
                                                          y))
                              (setdist (+ (first perm)
                                          (* (abs (- (length x) (length y))) PARAM-weight-length-difference)))
                              (lpgen (* (expt (+ (- dlen xi) 1);; log probability of generalizing
                                              PARAM-llBeta) 
                                        (- setdist)))
                              (pgen (max 0.001 (min 0.999 (exp lpgen)))) ;; prob of generalizing to this set
                              (alignment (cdr perm)))
                         
                            (for-each (lambda (j)
                                        (if (not (null? (nth j alignment))) ;; get nulls when x is longer than y
                                            (let* ((rj (first (nth j alignment))) ;; get the response to rj
                                                   (V  (if rj respn-T respn-F)))
                                              
                                             ; (displayn setdist "\t" lpgen "\t" pgen)
                                              (vector-set! V
                                                           (- j 1)
                                                           (+ (vector-ref V (- j 1))
                                                              pgen))
                                              )))
                                      (seq 1 (length x) 1))
                            ))
                     
                   ;  (displayn (map (lambda (t f) (/ t (+ t f)))
                    ;      (vector->list respn-T)
                     ;     (vector->list respn-F)))
                     
                     ;; now we want to convert T and F into things we can use. 
                     (map (lambda (t f) (/ t (+ t f)))
                          (vector->list respn-T)
                          (vector->list respn-F)))))))
           
    
   (define model-entropy (lambda args 0))
  
  ;; ---------------------
  ;;; And done, so make this thing a class
  (lambda (message . args)
    (apply (case message 
             ((initialize-model) initialize-model) 
             ((get-initial-p) get-initial-p)
             ((get-proposers) get-proposers) 
             ((model-push-hypothesis!) model-push-hypothesis!) 
             ((model-set-correct!) model-set-correct!) 
             ((model-set-which-item!) model-set-which-item!) 
             ((model-set-data!) model-set-data!)
             ((model-prediction) model-prediction) 
             ((model-entropy) model-entropy)
             (else (die (string-append "Unknown Method! " (symbol->string message))))) 
           args)))

                       
