#! gsc-script

;; This is a cute variant where the grammar parameters are unnormalized and play the role of a dirichlet hyperparams
;; in RR


;; This is like PCFG-flat but it doesn't store 
;; the priors, meaning that they are recomputed each time with a grammar you pass in to get model-prediction
;; This will compute how many times each
;; For each (vector->list (hashtable-keys SFL-grammar))

;; This stores counts of each rule usage and BV costs for each hypothesis,
;; and can take in a list of vectors of unnormalized log probabilities, 
;; one for each nonterminal. 

;; TODO October 25 2010 
;; TODO: We should set this so that bound variables aren't given any prob mass when they are impossible
;; - We can prevent ourselves from overwriting the hypotheses every time by saving the old rule counts



;; (model-push-hypothesis! h resp) -- pushes the hypothesis h with respones resp onto data
;; (model-set-correct! d) -- a set of data
;; (model-set-data! d)
;; (model-set-which-item! w) -- a list saying which item we are on
;; (model-prediction p) -- takes a parameter value and computes the model predicted proportions (possibly with fancy caching)
;; (get-lower-bound) -- retrieve bounds on p
;; (get-upper-bound) 
;; (get-initial-p) -- what p we can start with?
;; (get-proprosers) -- proposal functions for each of the parameters
;; (clear-hypotheses!)

; http://community.schemewiki.org/?simple-object

; -----------------------------------------------------------------------------------------
;; Define across parameters (outside here, so that they are shared)
(define (set-model)
  (set! across-parameters (vector 3.5 0.5 0.5 -0.7 -0.7)) ;; start high temperature to decrease burn-in
  (set! across-proposers (vector (lambda (x) (positive-normal-proposal x 0.25)) ;; temperature
                                 (lambda (x) (beta-proposal x 300))
                                 (lambda (x) (beta-proposal x 300))
                                 (lambda (x) (negative-normal-proposal x 0.05))
                                 (lambda (x) (negative-normal-proposal x 0.05))))
  (set! make-model make-model:parameterized-PCFG))

; -----------------------------------------------------------------------------------------
;; Define the "model" class
(define (make-model:parameterized-PCFG)
  
  (define META-description "The rational rules model, with beta fit across concepts." )
  (define META-notes       "(Uses C external library for running-ll)." )
  
  (define hypotheses null) ;; a vector of hypotheses
  (define responses null) ;; a vector of hypotheses
  (define responses_cb null)
  (define priors null) ;; a vector of hypotheses
  (define correct null) ;; stored in collapsed format
  (define correct_cb null)
  (define which-item null)
  (define which-item_cb null)
  (define zipped-responses null)
  (define len 0)
  (define hyp-rule-counts null) ;; for each hypothesis, a list of lists for each nonterminal of how often each rule was used. 
  
  (define (initialize-model) 0)  
  (define (model-set-data! d) null)
  
  ;; These are versions that are correct, but below may not use these since it caches
  (define (model-compute-prior p gp ap h) 
    (if (>= (SFL-grammar 'get-depth h) MAX-DEPTH)
        -inf.0
        (model-compute-prior-cnt (SFL-grammar 'get-rule-counts h)
                                 gp 
                                 (vector-ref ap 0))))
  
  (define (model-compute-prior-cnt cnt gp temperature)
    (/ (SFL-grammar 'log-probability-scorer RR-parameterized-scorer cnt gp)
       temperature))
  
  ;; this provides a wrapper for the C function that takes an N (index into data)
  ;; and a hypothesis 
  (define (model-compute-likelihood p gp ap D H)
    (let* ((PARAM-prior-temperature (vector-ref ap 0)) ;; extract these from across parameters
           (PARAM-alpha             (vector-ref ap 1))
           (PARAM-baserate          (vector-ref ap 2))
           (PARAM-llBeta-positive   (vector-ref ap 3))
           (PARAM-llBeta-negative   (vector-ref ap 4)))
      (compute-likelihood PARAM-alpha PARAM-baserate PARAM-llBeta-positive PARAM-llBeta-negative D H)))
  
  
  ;(compute-likelihood PARAM-alpha PARAM-baserate PARAM-llBeta D H)))
  
  (define (get-initial-p) (vector))
  (define (get-proposers) (vector))
  
  (define (model-set-correct! c) 
    (set! correct c)
    (set! len (length correct))
    (set! correct_cb (list->cblock (map to01 correct) 'int))
    ;(edisplayn len "\t" c)
    )
  
  (define (model-set-which-item! w) 
    (set! which-item w)
    ;(edisplayn w)
    (set! which-item_cb (list->cblock which-item 'int))
    )
  
  ;; adds h to the model, with c and l
  (define (model-push-hypothesis! h resp)
    (set! hypotheses (cons h hypotheses ))
    (set! responses  (cons resp responses ))
    (set! zipped-responses (zip-list responses)) ;; inefficient since we compute so many times!
    ;(edisplayn responses)
    (set! responses_cb (cons (list->cblock (map to01 resp) 'int) responses_cb))
    (set! hyp-rule-counts (cons (SFL-grammar 'get-rule-counts h)
                                hyp-rule-counts))
    
    )
  
  (define (clear-hypotheses!)
    (set! hypotheses null)
    (set! responses null)
    
    (for-each c_free responses_cb)
    (set! responses_cb null)
    (set! hyp-rule-counts null))
  
  ;; --------------------------
  ;; Local helper functions
  
  
  ;; get the posteriors -- the posterior prob for each hypothesis, at each data point
  ;; this is normalized!
  ;; This function needs to be here since it's how we interface sometimes. But caching the priors
  ;; speeds things up A LOT
  (define (get-posteriors p gp ap)
    (get-posteriors-prior p 
                          gp 
                          ap 
                          (map (lambda (hrc) (model-compute-prior-cnt hrc gp (vector-ref ap 0)))  ;; for each hypothesis, use hyp-rule-counts to compute the log probability according to grammar-prior
                               
                               hyp-rule-counts)))
  
  
  (define (get-posteriors-prior p gp ap priors)
    (let* ((PARAM-prior-temperature (vector-ref ap 0)) ;; extract these from across parameters
           (PARAM-alpha             (vector-ref ap 1))
           (PARAM-baserate          (vector-ref ap 2))
           (PARAM-llBeta-positive   (vector-ref ap 3))
           (PARAM-llBeta-negative   (vector-ref ap 4))
           ( running (map (lambda (rscb) (running-ll-C PARAM-alpha  ; (a b llb len resp_cb cor_cb which_cb)
                                                       PARAM-baserate
                                                       PARAM-llBeta-positive
                                                       PARAM-llBeta-negative
                                                       len
                                                       rscb
                                                       correct_cb
                                                       which-item_cb))
                          responses_cb)))
      
      (map (lambda (running-i)
             (let* ((unnormalized-post (map + running-i priors))
                    (Z (logsumexp-lazy unnormalized-post)))
               (map (lambda (y) (- y Z)) unnormalized-post)))
           (zip-list running))))
  
  ;; --------------------------
  ;; Here: grammar-prior is in the form of what (get-PCFG-rule-counts grammar t) gives back to you
  ;; so a list, one element per nonterminal, with the first element the prob of bound variables (which may be zero)
  ;; TODO: We should set this so that bound variables aren't given any prob mass when they are impossible
  (define (model-prediction p gp ap)
    ;; we compute priors out here since it is MUCH Faster
    (let ((priors (map (lambda (hrc) (model-compute-prior-cnt hrc gp (vector-ref ap 0)))
                       hyp-rule-counts)))
      
      (map (lambda (response-i posteriors) ;; responses for the i'th data point and the running-ll for the i'th data point
             (apply + (map (lambda (rij postj) 
                             (if (eqv? rij #t) ;; prob mass assigned to 1
                                 (exp postj)
                                 0.0))
                           response-i
                           posteriors)))
           zipped-responses ;; these in the right format--first list is all hyps on the first data point
           (get-posteriors-prior p gp ap priors)
           )))
  
  

  
  
  ;; compute the model entropy at each point
  (define (model-entropy p gp ap)
    (map (lambda (posteriors) ;; for each data point, this gives the posteriors for each hypothesis. 
           (- (/ (apply + (map (lambda (lp) (* (exp lp) lp)) ;; p*logp
                            posteriors))
                 0.6931471805599453))) ;; give it in bits
         (get-posteriors p gp ap)))
  
  ;; ------------------------------------------------
  ;;; And done, so make this thing a class
  (lambda (message . args)
    (apply (case message 
             ((initialize-model) initialize-model) 
             ((get-initial-p) get-initial-p)
             ((model-compute-prior) model-compute-prior)
             ((model-compute-likelihood) model-compute-likelihood)
             ((get-proposers) get-proposers) 
             ((model-push-hypothesis!) model-push-hypothesis!) 
             ((model-set-correct!) model-set-correct!) 
             ((model-set-which-item!) model-set-which-item!) 
             ((model-set-data!) model-set-data!)
             ((model-prediction) model-prediction) 
             ((model-entropy) model-entropy)
             ((clear-hypotheses!) clear-hypotheses!)
             (else (die (string-append "Unknown Method! " (symbol->string message))))) 
           args))) 


