

;; this is a hierarchical poisson model: Choose a number of primitives from 
;; a poisson, then choose uniformly from things with that number of primitives

;; (model-push-hypothesis! h resp) -- pushes the hypothesis h with respones resp onto data
;; (model-set-correct! d) -- a set of data
;; (model-set-data! d)
;; (model-set-which-item! w) -- a list saying which item we are on
;; (model-prediction p) -- takes a parameter value and computes the model predicted proportions (possibly with fancy caching)
;; (get-lower-bound) -- retrieve bounds on p
;; (get-upper-bound) 
;; (get-initial-p) -- what p we can start with?
;; (get-proprosers) -- proposal functions for each of the parameters


; -----------------------------------------------------------------------------------------
;; Define across parameters (outside here, so that they are shared)

(define (set-model)
  (set! across-parameters (vector  1.0 5.0 0.5 0.5 -0.5 -0.5))
  (set! across-proposers (vector
                          (lambda (x) (positive-normal-proposal x 0.2)) ;; temperature
                          (lambda (x) (positive-normal-proposal x 0.25)) 
                          (lambda (x) (beta-proposal x 200))
                          (lambda (x) (beta-proposal x 200))
                          (lambda (x) (negative-normal-proposal x 0.025))
                          (lambda (x) (negative-normal-proposal x 0.025))))
  (set! make-model make-model:poisson-length))

; -----------------------------------------------------------------------------------------
(define (make-model:poisson-length)
  
  (define META-description "Choose a length from poisson, and then uniformly from those of that length." )
  (define META-notes       "(Uses C external library for running-ll)." )
  
  (define hypotheses null) ;; a vector of hypotheses
  (define responses null) ;; a vector of hypotheses
  (define responses_cb null) ;; a vector of hypotheses
  (define priors null) ;; a vector of hypotheses
  (define correct null) ;; stored in collapsed format
  (define correct_cb null) ;; stored in collapsed format
  (define which-item null)
  (define which-item_cb null) ;; stored in collapsed format
  (define length2count (make-vector 100 0)) ;; this maps each length (number of primitives) to a count of how many
  (define len 0)
  
  
  
  ;; this fills up length2count
  (define (initialize-model) 
    (for h in hypotheses
         (let ((lh (hypothesis-length h)))
           (vector-set! length2count 
                        lh
                        (+ 1 (vector-ref length2count lh))))))
    
  (define (model-set-data! d) null)
  (define (get-initial-p) (vector))
  (define (get-proposers) (vector))
                           
                           
 (define (model-set-correct! c) 
    (set! correct c)
    (set! len (length correct))
    (set! correct_cb (list->cblock (map to01 correct) 'int))
    )
  
  (define (model-set-which-item! w) 
    (set! which-item w)
    (set! which-item_cb (list->cblock which-item 'int))
    )
  
  ;; adds h to the model, with c and l
  (define (model-push-hypothesis! h resp) 
    (set! hypotheses (cons h hypotheses ))
    (set! responses  (cons resp responses ))
    (set! responses_cb (cons (list->cblock (map to01 resp) 'int) responses_cb))
    )
   
  (define (clear-hypotheses!) null)
  ;; --------------------------
  ;; Local helper functions
  
  ;; for now, we take length as the number of nodes (instead of num of funcs, etc)
  (define (hypothesis-length t) (SFL-grammar 'count-nodes t))
    
  ;; --------------------------
  (define (model-prediction p gp ap)
    
    (let* ((PARAM-prior-temperature (vector-ref ap 0))
           (PARAM-prior-lambda      (vector-ref ap 1))
           (PARAM-alpha             (vector-ref ap 2))
           (PARAM-baserate          (vector-ref ap 3))
           
           (PARAM-llBeta-positive            (vector-ref ap 4))
           (PARAM-llBeta-negative            (vector-ref ap 5))
           
           ;; do this so we can use the same below code
           ;; The running LL with the decay
           
           (running (map (lambda (rs) (let* ((rscb (list->cblock (map to01 rs) 'int))
                                             (ret   (running-ll-C PARAM-alpha  ; (a b llb len resp_cb cor_cb which_cb)
                                                                  PARAM-baserate
                                                                  PARAM-llBeta-positive
                                                                  PARAM-llBeta-negative
                                                                  len
                                                                  rscb
                                                                  correct_cb
                                                                  which-item_cb)))
                                        (free rscb)
                                        ret))
                         responses))
           )
      
      (map (lambda (response-i running-i) ;; the responses for the i'th data point and the running-ll for the i'th data point
             (let* ((posterior (map (lambda (ri h)
                                      (let ((hl (hypothesis-length h)))
                                        (+ ri 
                                           (/ 1.0 PARAM-prior-temperature)
                                           (log (/ 1.0 (vector-ref length2count hl))) ;; choose uniformly from this length
                                           (* (log PARAM-prior-lambda) hl) ;; poisson prior on length
                                           (- PARAM-prior-lambda)
                                           (- (gsl-lngamma (+ 1 hl))) ;; +1 since its the gamma function
                                           )))
                                    running-i
                                    hypotheses))
                    (normalizer (logsumexp-lazy posterior)))
               (sum (map (lambda (rij postj) 
                           (* (if (eqv? rij #t) 1.0 0.0) 
                              (exp (- postj normalizer))))
                         response-i
                         posterior))))
           (zip-list responses) ;; these in the right format--first list is all hyps on the first data point
           (zip-list running)
           )))
  
  (define model-entropy (lambda args 0))
  
  ;; ------------------------------------------------
  ;;; And done, so make this thing a class
  (lambda (message . args)
    (apply (case message 
             ((initialize-model) initialize-model) 
             ((get-initial-p) get-initial-p)
             ((get-proposers) get-proposers) 
             ((model-push-hypothesis!) model-push-hypothesis!) 
             ((model-set-correct!) model-set-correct!) 
             ((model-set-which-item!) model-set-which-item!) 
             ((model-set-data!) model-set-data!)
             ((model-prediction) model-prediction) 
	     ((clear-hypotheses!) clear-hypotheses!)
             ((model-entropy) model-entropy)
             (else (die (string-append "Unknown Method! " (symbol->string message))))) 
           args))) 

                       
